"""toil_container containers tests."""

from multiprocessing import Process
from os.path import join
import getpass
import os

import docker
import pytest

from toil_container import __version__
from toil_container import exceptions
from toil_container.containers import _TMP_PREFIX
from toil_container.containers import _remove_docker_container
from toil_container.containers import docker_call
from toil_container.containers import singularity_call

from .utils import DOCKER_IMAGE
from .utils import ROOT
from .utils import SINGULARITY_IMAGE
from .utils import SKIP_DOCKER
from .utils import SKIP_SINGULARITY
from .utils import Capturing


def assert_option_check_output(call, img):
    # test check_output True
    assert "bin" in call(img, args=["ls", "/"], check_output=True)

    # test toil_container.ContainerError is raised with bad command
    with pytest.raises(exceptions.ContainerError) as error:
        call(img, args=["rm", "/bin"])

    assert "raised during the container system call" in str(error.value)


def assert_option_cwd(call, img):
    assert "bin" in call(img, ["ls", ".."], cwd="/bin", check_output=True)


def assert_option_env(call, img):
    args = ["bash", "-c", "echo $FOO"]
    assert "BAR" in call(img, args, env={"FOO": "BAR"}, check_output=True)

    # check container doesn't inherit environment
    os.environ["FOO"] = "BAR"
    args = ["bash", "-c", "echo $FOO"]
    assert "BAR" not in call(img, args, check_output=True)


def assert_parallel_call(call, img):
    p1 = Process(target=lambda: call(img, ["sleep", "1"]))
    p2 = Process(target=lambda: call(img, ["sleep", "1"]))
    p1.start()
    p2.start()
    p1.join()
    p2.join()


def assert_option_volumes(call, img, tmpdir):
    vsrc = tmpdir.strpath
    fsrc = tmpdir.join("foo")
    vdst = join(os.sep, "SHARED")
    fdst = join(vdst, "foo")
    call(img, ["bash", "-c", "echo bar > " + fdst], volumes=[(vsrc, vdst)])
    assert "bar" in fsrc.read()


def assert_option_working_dir(call, img, tmpdir):
    args = ["bash", "-c", "echo bar > /tmp/foo"]
    dont_remove = tmpdir.mkdir("dont")
    call(img, args, working_dir=dont_remove.strpath, remove_tmp_dir=False)
    tmpfile = next(dont_remove.visit(_TMP_PREFIX + "*/foo"))
    assert "bar" in tmpfile.read()

    remove = tmpdir.mkdir("remove")
    call(img, args, working_dir=remove.strpath, remove_tmp_dir=True)
    assert not list(remove.visit(_TMP_PREFIX + "*"))


@SKIP_DOCKER
def test_docker_check_output_false_prints_to_stdout_and_stderr():
    with Capturing() as output:
        exit_status = docker_call(DOCKER_IMAGE, args=["ls", "/"])

    assert exit_status == 0
    assert "bin" in " ".join(output)  # check stdout and stderr are printed


@SKIP_DOCKER
def test_docker_check_output():
    assert_option_check_output(docker_call, DOCKER_IMAGE)


@SKIP_DOCKER
def test_docker_cwd():
    assert_option_cwd(docker_call, DOCKER_IMAGE)


@SKIP_DOCKER
def test_docker_env():
    assert_option_env(docker_call, DOCKER_IMAGE)


@SKIP_DOCKER
def test_docker_parallel():
    assert_parallel_call(docker_call, DOCKER_IMAGE)


@SKIP_DOCKER
def test_docker_volumes(tmpdir):
    assert_option_volumes(docker_call, DOCKER_IMAGE, tmpdir)


@SKIP_DOCKER
def test_docker_working_dir(tmpdir):
    assert_option_working_dir(docker_call, DOCKER_IMAGE, tmpdir)


@SKIP_SINGULARITY
def test_singularity_check_output():
    assert_option_check_output(singularity_call, SINGULARITY_IMAGE)


@SKIP_SINGULARITY
def test_singularity_cwd():
    assert_option_cwd(singularity_call, SINGULARITY_IMAGE)


@SKIP_SINGULARITY
def test_singularity_env():
    assert_option_env(singularity_call, SINGULARITY_IMAGE)


@SKIP_SINGULARITY
def test_singularity_parallel():
    assert_parallel_call(singularity_call, SINGULARITY_IMAGE)


@SKIP_SINGULARITY
def test_singularity_volumes(tmpdir):
    assert_option_volumes(singularity_call, SINGULARITY_IMAGE, tmpdir)


@SKIP_SINGULARITY
def test_singularity_working_dir(tmpdir):
    assert_option_working_dir(singularity_call, SINGULARITY_IMAGE, tmpdir)


@SKIP_SINGULARITY
def test_singularity_doesnt_overwrite_home():
    args = ["bash", "-c", "ls /home"]
    skip = ".singularity/docker"
    output = singularity_call(SINGULARITY_IMAGE, args, check_output=True)
    output = "".join(i for i in output.split() if skip not in i)
    assert getpass.getuser() not in output


@SKIP_DOCKER
def test_remove_docker_container():
    name = "florentino-ariza"
    client = docker.from_env(version="auto")
    container = client.containers.create(DOCKER_IMAGE, ["ls"], name=name)
    container.start()
    _remove_docker_container(name)

    with pytest.raises(docker.errors.NotFound) as error:
        client.containers.get(name)

    assert name in str(error.value)


@SKIP_DOCKER
def test_docker_container():
    python_args = "from toil_container import __version__; print(__version__)"
    args = ["python", "-c", python_args]
    image_tag = "test-toil-container-image"
    client = docker.from_env(version="auto")
    client.images.build(path=ROOT, rm=True, tag=image_tag)
    assert __version__ in docker_call(image_tag, args, check_output=True)



"""toil_container tests utils."""

from cStringIO import StringIO
from os.path import abspath
from os.path import dirname
from os.path import join
import os
import sys

import pytest

from toil_container import utils

ROOT = abspath(join(dirname(__file__), ".."))

SKIP_DOCKER = pytest.mark.skipif(
    not utils.is_docker_available(),
    reason="docker is not available."
    )

SKIP_SINGULARITY = pytest.mark.skipif(
    not utils.is_singularity_available(),
    reason="singularity is not available."
    )

DOCKER_IMAGE = "ubuntu:latest"

if os.path.isfile(os.getenv("CACHED_SINGULARITY_IMAGE", "/")):
    SINGULARITY_IMAGE = os.getenv("CACHED_SINGULARITY_IMAGE")
else:
    SINGULARITY_IMAGE = "docker://" + DOCKER_IMAGE


class Capturing(list):

    """
    Capture stdout and stderr of a function call.

    See: https://stackoverflow.com/questions/16571150

    Example:

        with Capturing() as output:
            do_something(my_object)
    """

    def __enter__(self):
        self._stdout = sys.stdout
        self._stderr = sys.stderr
        sys.stderr = sys.stdout = self._stringio = StringIO()
        return self

    def __exit__(self, *args):
        self.extend(self._stringio.getvalue().splitlines())
        del self._stringio  # free up some memory
        sys.stdout = self._stdout
        sys.stderr = self._stderr



"""toil_container jobs tests."""

import argparse

import pytest

from toil_container import exceptions
from toil_container import jobs

from .utils import DOCKER_IMAGE
from .utils import SINGULARITY_IMAGE
from .utils import SKIP_DOCKER
from .utils import SKIP_SINGULARITY


def test_call_uses_subprocess():
    options = argparse.Namespace()
    job = jobs.ContainerJob(options)
    assert job.call(["ls"]) == 0
    assert "bin" in job.call(["ls", "/"], check_output=True)

    # check subprocess.CalledProcessError
    with pytest.raises(exceptions.SystemCallError):
        job.call(["rm", "/bin"])

    # check OSError
    with pytest.raises(exceptions.SystemCallError):
        job.call(["florentino-ariza"])


def assert_image_call(image_attribute, image, tmpdir):
    """Get options namespace."""
    options = argparse.Namespace()
    options.workDir = tmpdir.mkdir("working_dir").strpath
    setattr(options, image_attribute, image)

    # create job and options
    vol1 = tmpdir.mkdir("vol1")
    vol2 = tmpdir.mkdir("vol2")
    options.volumes = [
        (vol1.strpath, "/vol1"),
        (vol2.strpath, "/vol2"),
        ]

    vol1.join("foo").write("bar")
    job = jobs.ContainerJob(options)

    # test cwd
    assert "bin" in job.call(["ls", ".."], cwd="/bin", check_output=True)

    # test volume
    assert "bar" in job.call(["cat", "/vol1/foo"], check_output=True)
    assert job.call(["ls"]) == 0

    # test env
    cmd = ["bash", "-c", "echo $FOO"]
    assert "BAR" in job.call(cmd, env={"FOO": "BAR"}, check_output=True)

    # check subprocess.CalledProcessError
    with pytest.raises(exceptions.SystemCallError):
        job.call(["rm", "/bin"])

    # check OSError
    with pytest.raises(exceptions.SystemCallError):
        job.call(["florentino-ariza"])

    # test both singularity and docker raiser error
    options = argparse.Namespace()
    options.docker = "foo"
    options.singularity = "bar"
    job = jobs.ContainerJob(options)

    with pytest.raises(exceptions.UsageError) as error:
        job.call(["foo", "bar"])

    assert "Both docker and singularity can't be set" in str(error.value)


@SKIP_DOCKER
def test_job_with_docker_call(tmpdir):
    assert_image_call("docker", DOCKER_IMAGE, tmpdir)


@SKIP_SINGULARITY
def test_job_with_singularity_call(tmpdir):
    assert_image_call("singularity", SINGULARITY_IMAGE, tmpdir)



"""toil_container utils."""

import os
import sys

# learn more here: https://github.com/google/python-subprocess32
if os.name == "posix" and sys.version_info[0] < 3:
    import subprocess32 as subprocess
else:
    import subprocess

import docker
import requests

from toil_container import exceptions


def is_docker_available(raise_error=False, path=False):
    """
    Check if docker is available to run in the current environment.

    Arguments:
        raise_error (bool): flag to raise error when command is unavailable.
        path (bool): flag to return location of the command in the user's path.

    Returns:
        bool: True if docker is available.

    Raises:
        OSError: if the raise_error flag was passed as an argument and the
        command is not available to execute.
    """
    expected_exceptions = (
        requests.exceptions.ConnectionError,
        docker.errors.APIError
    )

    try:
        # Test docker is running
        client = docker.from_env()
        is_available = client.ping()

        if path:
            return which("docker")
        return is_available

    except expected_exceptions as error:
        if raise_error:
            raise exceptions.DockerNotAvailableError(str(error))
        return False


def is_singularity_available(raise_error=False, path=False):
    """
    Check if singularity is available to run in the current environment.

    Arguments:
        raise_error (bool): flag to raise error when command is unavailable.
        path (bool): flag to return location of the command in the user's path.

    Returns:
        bool: True if singularity is available.
        str: absolute location of the file in the user's path.

    Raises:
        OSError: if the raise_error flag was passed as an argument and the
        command is not available to execute.
    """
    try:
        subprocess.check_output(["singularity", "--version"])

        if path:
            return which('singularity')
        return True

    except (subprocess.CalledProcessError, OSError) as error:
        if raise_error:
            raise exceptions.SingularityNotAvailableError(str(error))
        return False


def get_container_error(error):
    """Return a ContainerError with information about `error`."""
    return exceptions.ContainerError(
        "The following error was raised during the container system call: "
        "{}: {}".format(type(error), str(error))
        )


def which(program):
    """
    Locate a program file in the user's path.

    Python implementation to mimic the behavior of the UNIX 'which' command
    And shutil.which() is not supported in python 2.x.

    See: https://stackoverflow.com/questions/377017

    Arguments:
        program (str): command to be tested. Can be relative or absolute path.

    Return:
        str: program file in the user's path.
    """
    fpath, _ = os.path.split(program)
    if fpath:
        if _is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            exe_file = os.path.join(path, program)
            if _is_exe(exe_file):
                return exe_file
    return None


def _is_exe(fpath):
    """
    Check if fpath is executable for the current user.

    Arguments:
        fpath (str): relative or absolute path.

    Return:
        bool: True if execution is granted, else False.
    """
    return os.path.isfile(fpath) and os.access(fpath, os.X_OK)



"""toil_container version test."""

import subprocess

from toil_container import utils


def test_which_util():
    """Test to check the which util is working."""
    std_out = subprocess.check_output(
        [utils.which("python"), "--version"],
        stderr=subprocess.STDOUT
        )
    assert "2.7" in std_out.decode()



"""toil_container specific exceptions."""


class ToilContainerException(Exception):

    """A base exception for toil_container."""


class UsageError(ToilContainerException):

    """A class to raise when improper usage."""


class ContainerError(ToilContainerException):

    """A class to raise when a container call fails."""


class SystemCallError(ToilContainerException):

    """A class to raise when a system call cannot be completed."""


class ValidationError(ToilContainerException):

    """A class to raise for validation errors."""


class ToolNotAvailableError(ToilContainerException):

    """A base exception to raise when tools are not available."""


class DockerNotAvailableError(ToolNotAvailableError):

    """A class to raise when docker is not available."""


class SingularityNotAvailableError(ToolNotAvailableError):

    """A class to raise when singularity is not available."""



"""toil_container validators."""

from toil_container import exceptions
from toil_container.containers import docker_call
from toil_container.containers import singularity_call


def validate_docker(image, volumes=None, working_dir=None):
    """Validate a docker image."""
    _validate_image(docker_call, image, volumes, working_dir)
    return image


def validate_singularity(image, volumes=None, working_dir=None):
    """Validate a singularity image."""
    _validate_image(singularity_call, image, volumes, working_dir)
    return image


def _validate_image(call, image, volumes, working_dir):
    """Call will fail if invalid image, volumes or working_dir are passed."""
    cmd = ["bash", "-c"]

    if volumes:
        cmd += [" && ".join(["ls " + i[1] for i in volumes])]
    else:
        cmd += ["echo"]

    try:
        kwargs = dict(volumes=volumes, working_dir=working_dir)
        call(image, cmd, check_output=True, **kwargs)
    except exceptions.ContainerError as error:
        raise exceptions.ValidationError(
            "Invalid container configuration: "
            "{}: {}".format(type(error), str(error))
            )



"""toil_container pasers."""

import argparse

from toil.job import Job
import click

from toil_container import validators

CUSTOM_TOIL_ACTIONS = [
    argparse.Action(
        ["TOIL OPTIONAL ARGS"],
        dest="",
        default=argparse.SUPPRESS,
        help="see --help-toil for a full list of toil parameters",
        ),
    argparse.Action(
        [],
        dest="jobStore",
        help="the location of the job store for the workflow. "
        "See --help-toil for more information [REQUIRED]",
        ),
    ]


class _ToilHelpAction(argparse._HelpAction):

    def __call__(self, parser, namespace, values, option_string=None):
        """Print parser help and exist whilst adding a flag to the parser."""
        parser.show_toil_groups = True
        parser.print_help()
        parser.exit()


class ToilBaseArgumentParser(argparse.ArgumentParser):

    """Add toil options to argument parser."""

    def __init__(self, version=None, **kwargs):
        """
        Add Toil options to parser.

        Arguments:
            version (str): optionally add a version argument.
            kwargs (dict): argparse.ArgumentParser key word arguments.
        """
        if not kwargs.get("formatter_class"):
            kwargs["formatter_class"] = argparse.ArgumentDefaultsHelpFormatter

        super(ToilBaseArgumentParser, self).__init__(**kwargs)

        if version:
            self.add_argument(
                "-v", "--version",
                action="version",
                version="%(prog)s " + str(version)
                )

        Job.Runner.addToilOptions(self)


class ToilShortArgumentParser(ToilBaseArgumentParser):

    """
    A parser with only required toil args in --help with --help-toil option.

    Toil options are automatically added, but hidden by default in the help
    print. However, the `--help-toil` argument prints toil full rocketry.
    """

    def __init__(self, **kwargs):
        """Add Toil`--help-toil` to parser."""
        super(ToilShortArgumentParser, self).__init__(**kwargs)

        self.add_argument(
            "--help-toil",
            action=_ToilHelpAction, default=argparse.SUPPRESS,
            help="print help with full list of Toil arguments and exit"
            )

    def get_help_groups(self, show_toil_groups):
        """Decide whether to show toil options or not."""
        action_groups = []
        actions = []

        for action_group in self._action_groups:
            is_toil_group = action_group.title.startswith("toil")
            is_toil_group |= "Logging Options" in action_group.title

            if not is_toil_group or (is_toil_group and show_toil_groups):
                action_groups.append(action_group)
                actions += action_group._group_actions

        return actions, action_groups

    def format_help(self):
        """Include toil options if `self.show_toil_groups` is True."""
        formatter = self._get_formatter()

        # decide whether to show toil options or not
        show_toil_groups = getattr(self, "show_toil_groups", False)
        actions, action_groups = self.get_help_groups(show_toil_groups)

        # usage, the CUSTOM_TOIL_ACTIONS are just for display
        formatter.add_usage(
            self.usage,
            actions + ([] if show_toil_groups else CUSTOM_TOIL_ACTIONS),
            self._mutually_exclusive_groups
            )

        # description
        formatter.add_text(self.description)

        # positionals, optionals and user-defined groups
        for action_group in action_groups:
            formatter.start_section(action_group.title)
            formatter.add_text(action_group.description)
            formatter.add_arguments(action_group._group_actions)
            formatter.end_section()

        # add custom toil section
        if not show_toil_groups:
            formatter.start_section("toil arguments")
            formatter.add_arguments(CUSTOM_TOIL_ACTIONS)
            formatter.end_section()

        # epilog
        formatter.add_text(self.epilog)

        # determine help from format above
        return formatter.format_help()


class ContainerArgumentParser(ToilShortArgumentParser):

    """Toil Argument Parser with options for containerized system calls."""

    def __init__(self, *args, **kwargs):
        """Add container options to parser."""
        super(ContainerArgumentParser, self).__init__(*args, **kwargs)
        settings = self.add_argument_group("container arguments")

        settings.add_argument(
            "--docker",
            help="name/path of the docker image available in daemon",
            default=None,
            required=False,
            )

        settings.add_argument(
            "--singularity",
            help="name/path of the singularity image available in deamon",
            default=None,
            required=False,
            )

        settings.add_argument(
            "--volumes",
            help="tuples of (local path, absolute container path)",
            required=False,
            default=None,
            action="append",
            nargs=2,
            )

    def parse_args(self, args=None, namespace=None):
        """Validate parsed options."""
        args = super(ContainerArgumentParser, self).parse_args(
            args=args, namespace=namespace
            )

        images = [args.docker, args.singularity]

        if all(images):
            raise click.UsageError(
                "You can't pass both --singularity and --docker."
                )

        if args.volumes and not any(images):
            raise click.UsageError(
                "--volumes should be used only with "
                "--singularity or --docker."
                )

        if any(images):
            validate_kwargs = {}

            if args.volumes:
                validate_kwargs["volumes"] = args.volumes

            if args.workDir:
                validate_kwargs["working_dir"] = args.workDir

            if args.docker:
                validate_kwargs["image"] = args.docker
                validators.validate_docker(**validate_kwargs)
            else:
                validate_kwargs["image"] = args.singularity
                validators.validate_singularity(**validate_kwargs)

        return args



"""
Module to manage docker and singularity calls through the containers tools.

Docker uses a API for python, while singularity API doesn't support python2.

Based on the docker implementation of:
https://github.com/BD2KGenomics/toil/blob/master/src/toil/lib/docker.py

Based on the singularity implementation of:
https://github.com/vgteam/toil-vg/blob/master/src/toil_vg/singularity.py
"""

from __future__ import print_function

from tempfile import mkdtemp
import os
import shutil
import sys
import uuid

import docker

from toil_container.utils import get_container_error
from toil_container.utils import is_docker_available
from toil_container.utils import is_singularity_available
from toil_container.utils import subprocess

_TMP_PREFIX = "toil_container_tmp_"


def singularity_call(
        image,
        args=None,
        cwd=None,
        env=None,
        check_output=None,
        working_dir=None,
        volumes=None,
        remove_tmp_dir=True):
    """
    Execute parameters in a singularity container via subprocess.

    Singularity will be called with the following command:

        singularity -q exec
            --bind <shared_fs>:<shared_fs>     # if shared_fs is provided
            --contain --workdir <working_dir>  # if working_dir is provided
            --pwd {cwd}                        # if cwd is provided
            <image> <args>

    Docker images can be run by prefacing the input image with 'docker://'.
    In this case, Singularity will download, convert, and cache the image
    on the fly. This cache can be set with SINGULARITY_CACHEDIR, and
    defaults to the user's home directory. This cache can be a major
    bottleneck when repeatedly more different images than it can hold
    (not very many). So for this type of usage pattern (concurrent or short
    consecutive calls to different images), it is best to run Singularity
    images natively.

    Arguments:
        image (str): name/path of the image.
        args (list): list of command line arguments passed to the tool.
        cwd (str): current working directory.
        env (dict): environment variables to set inside container.
        check_output (bool): check_output or check_call behavior.
        working_dir (str): path to a working directory. If passed, a tmpdir
            will be created inside and will be mounted to /tmp.
        volumes (list): list of tuples (src-path, dst-path) to be mounted,
            dst-path must be absolute path.
        remove_tmp_dir (bool): remove tmpdir created inside `working_dir`.

    Returns:
        str: (check_output=True) stdout of the system call.
        int: (check_output=False) 0 if call succeed else non-0.

    Raises:
        toil_container.ContainerError: if the container invocation fails.
        toil_container.SingularityNotAvailableError: singularity not installed.
    """
    singularity_path = is_singularity_available(raise_error=True, path=True)

    # ensure singularity doesn't overwrite $HOME by pointing to dummy dir
    # /tmp will be mapped to work_dir/scratch/tmp and removed after the call
    home_dir = ".unused_home"
    work_dir = mkdtemp(prefix=_TMP_PREFIX, dir=working_dir)
    os.makedirs(os.path.join(work_dir, "scratch", "tmp", home_dir))
    singularity_args = [
        "--scratch", "/tmp",
        "--home", "{}:/tmp/{}".format(os.getcwd(), home_dir),
        "--workdir", work_dir,
        ]

    # set parameters for managing directories if options are defined
    if volumes:
        for src, dst in volumes:
            singularity_args += ["--bind", "{}:{}".format(src, dst)]

    if cwd:
        singularity_args += ["--pwd", cwd]

    # setup the outgoing subprocess call for singularity
    command = [singularity_path, "-q", "exec"] + singularity_args
    command += [image] + (args or [])

    if check_output:
        call = subprocess.check_output
    else:
        call = subprocess.check_call

    try:
        output = call(command, env=env or {})
        error = False
    except (subprocess.CalledProcessError, OSError) as error:
        pass

    if remove_tmp_dir:
        try:
            shutil.rmtree(work_dir)
        except:  # pylint: disable=W0702
            pass

    if error:
        raise get_container_error(error)

    try:
        return output.decode()
    except AttributeError:
        return output


def docker_call(
        image,
        args=None,
        cwd=None,
        env=None,
        check_output=None,
        working_dir=None,
        volumes=None,
        remove_tmp_dir=True):
    """
    Execute parameters in a docker container via docker-python API.

    See: https://docker-py.readthedocs.io/en/stable/

    Arguments:
        image (str): name/path of the image.
        args (list): list of command line arguments passed to the tool.
        cwd (str): current working directory.
        env (dict): environment variables to set inside container.
        check_output (bool): check_output or check_call behavior.
        working_dir (str): path to a working directory. If passed, a tmpdir
            will be created inside and will be mounted to the container /tmp.
        volumes (list): list of tuples (src-path, dst-path) to be mounted,
            dst-path must be absolute path.
        remove_tmp_dir (bool): remove tmpdir created inside `working_dir`.

    Returns:
        str: (check_output=True) stdout of the system call.
        int: (check_output=False) 0 if call succeed else raise error.

    Raises:
        toil_container.ContainerError: if the container invocation fails.
        toil_container.DockerNotAvailableError: when docker not available.
    """
    is_docker_available(raise_error=True)
    container_name = "container-" + str(uuid.uuid4())
    work_dir = None
    kwargs = {}
    kwargs["command"] = args
    kwargs["entrypoint"] = ""
    kwargs["environment"] = env or {}
    kwargs["name"] = container_name
    kwargs["volumes"] = {}

    # Set parameters for managing directories if options are defined
    if volumes:
        for src, dst in volumes:
            kwargs["volumes"][src] = {"bind": dst, "mode": "rw"}

    if working_dir:
        # if working_dir is passed, we need to make sure it will be unique
        work_dir = mkdtemp(prefix=_TMP_PREFIX, dir=working_dir)
        kwargs["volumes"][work_dir] = {"bind": "/tmp", "mode": "rw"}

    if cwd:
        kwargs["working_dir"] = cwd

    client = docker.from_env(version="auto")
    expected_errors = (
        docker.errors.ImageNotFound,
        docker.errors.APIError,
        )

    try:
        container = client.containers.run(image, detach=True, **kwargs)
        exit_status = container.wait()
        error = False
    except expected_errors as error:
        pass

    if remove_tmp_dir:
        try:
            shutil.rmtree(work_dir)
        except:  # pylint: disable=W0702
            pass

    if error:
        _remove_docker_container(container_name)
        raise get_container_error(error)

    if check_output:
        output = container.logs(stdout=True, stderr=False)
        stderr = container.logs(stdout=True, stderr=True)

    else:  # print logs to stdout and stderr as subprocess.check_call
        output = exit_status
        stderr = container.logs(stdout=False, stderr=True)
        stdout = container.logs(stdout=True, stderr=False)
        print(stderr, file=sys.stderr)
        print(stdout, file=sys.stdout)

    container.stop()
    container.remove()

    if exit_status != 0:
        error = docker.errors.ContainerError(
            container=container,
            exit_status=exit_status,
            command=args,
            image=image,
            stderr=stderr,
            )

        raise get_container_error(error)

    try:
        return output.decode()
    except AttributeError:
        return output


def _remove_docker_container(container_name):
    try:
        client = docker.from_env(version="auto")
        container = client.containers.get(container_name)
        container.stop()
        container.remove()
    except docker.errors.APIError:
        pass



"""toil_container integration tests."""

from os.path import join

from toil_container import jobs
from toil_container import parsers
from toil_container.containers import _TMP_PREFIX

from .utils import DOCKER_IMAGE
from .utils import SINGULARITY_IMAGE
from .utils import SKIP_DOCKER
from .utils import SKIP_SINGULARITY


class ContainerTestJob(jobs.ContainerJob):

    cmd = ["pwd"]
    cwd = None
    env = {}
    check_output = True

    def run(self, jobStore):
        self.call(
            self.cmd,
            cwd=self.cwd,
            env=self.env,
            check_output=self.check_output
            )


def assert_pipeline(image_flag, image, tmpdir):
    """
    Make sure parallel jobs work.

                head
                |  |
          child_a  child_b
                |  |
                tail
    """
    jobstore = tmpdir.join("jobstore")
    workdir = tmpdir.mkdir("working_dir")
    local_volume = tmpdir.mkdir("volume")
    container_volume = "/volume"

    # set parser
    parser = parsers.ContainerArgumentParser()
    args = [jobstore.strpath, "--workDir", workdir.strpath]

    if image_flag:
        args += [
            image_flag, image,
            "--volumes", local_volume.strpath, container_volume,
            ]

    # set testing variables
    out_file = "bottle.txt"
    tmp_file_local = workdir.join(out_file)
    tmp_file_container = join("/tmp", out_file)
    vol_file_local = local_volume.join(out_file)
    vol_file_container = join(container_volume, out_file)

    if image_flag is None:
        tmp_file_container = tmp_file_local.strpath
        vol_file_container = vol_file_local.strpath

    # create jobs
    options = parser.parse_args(args)
    head = ContainerTestJob(options)
    child_a = ContainerTestJob(options)
    child_b = ContainerTestJob(options)
    tail = ContainerTestJob(options)

    # assign commands and attributes
    cmd = ["/bin/bash", "-c"]

    # test cwd and workDir, _rm_tmp_dir is used to prevent tmpdir to be removed
    head._rm_tmp_dir = False
    head.cwd = "/bin"
    head.cmd = cmd + ["pwd >> " + tmp_file_container]

    # test env
    child_a.env = {"FOO": "BAR"}
    child_a.cmd = cmd + ["echo $FOO >> " + vol_file_container]

    # test check_output
    child_b.check_output = False
    child_b.cmd = cmd + ["echo check_call >> " + vol_file_container]

    # test volumes
    tail.cmd = cmd + ["echo volume >> " + vol_file_container]

    # build dag
    head.addChild(child_a)
    head.addChild(child_b)
    head.addFollowOn(tail)

    # start pipeline
    jobs.ContainerJob.Runner.startToil(head, options)

    if image_flag:
        pattern = join(_TMP_PREFIX + "*", out_file)
        tmp_file_local = next(workdir.visit(pattern))

    # Test the output
    with open(tmp_file_local.strpath) as f:
        result = f.read()
        assert "/bin" in result

    if image_flag:
        with open(vol_file_local.strpath) as f:
            result = f.read()
            assert "volume" in result
            assert "BAR" in result
            assert "check_call" in result


def test_pipeline_with_subprocess(tmpdir):
    assert_pipeline(None, None, tmpdir)


@SKIP_DOCKER
def test_pipeline_with_docker(tmpdir):
    assert_pipeline("--docker", DOCKER_IMAGE, tmpdir)


@SKIP_SINGULARITY
def test_pipeline_with_singularity(tmpdir):
    assert_pipeline("--singularity", SINGULARITY_IMAGE, tmpdir)



"""toil_container module."""

from os.path import abspath
from os.path import dirname
from os.path import join

from toil_container.containers import (
    docker_call,
    singularity_call,
    )

from toil_container.jobs import (
    ContainerJob,
    )

from toil_container.parsers import (
    ContainerArgumentParser,
    ToilShortArgumentParser,
    )

from toil_container.exceptions import (
    ContainerError,
    DockerNotAvailableError,
    SingularityNotAvailableError,
    ToilContainerException,
    ToolNotAvailableError,
    UsageError,
    )

# make sure we use absolute paths
ROOT = abspath(dirname(__file__))

with open(join(ROOT, "VERSION"), "r") as f:
    VERSION = f.read().strip()

__version__ = VERSION



"""toil_container setup.py."""

from os.path import abspath
from os.path import dirname
from os.path import join
import json

from setuptools import find_packages
from setuptools import setup

# make sure we use absolute paths
ROOT = abspath(dirname(__file__))

# please put setup keywords in the setup.json to keep this file clean
with open(join(ROOT, "setup.json"), "r") as f:
    SETUP = json.load(f)

# see 4 > https://packaging.python.org/guides/single-sourcing-package-version/
with open(join(ROOT, "toil_container", "VERSION"), "r") as f:
    VERSION = f.read().strip()

setup(
    # single source package version
    version=VERSION,

    # in combination with recursive-includes in MANIFEST.in, non-python files
    # within the toil_container will be copied into the site-packages and
    # wheels installation directories
    include_package_data=True,

    # return a list all Python packages found within the ROOT directory
    packages=find_packages(),

    # pass parameters loaded from setup.json including author and version
    **SETUP
)



"""toil_container jobs."""

from toil.job import Job

from toil_container import containers
from toil_container import exceptions
from toil_container.utils import subprocess


class ContainerJob(Job):

    """A job class with a `call` method for containerized system calls."""

    def __init__(self, options, *args, **kwargs):
        """
        Set toil's namespace `options` as an attribute.

        Use toil_container.

        Arguments:
            options (object): an `argparse.Namespace` object with toil options.
            args (list): positional arguments to be passed to `toil.job.Job`.
            kwargs (dict): key word arguments to be passed to `toil.job.Job`.
        """
        self.options = options
        super(ContainerJob, self).__init__(*args, **kwargs)

    def call(self, args, cwd=None, env=None, check_output=False):
        """
        Make a containerized call if images available, else use subprocess.

        Docker will be used if `self.options.docker` is set. Similarly,
        Singularity will be used if `self.options.singularity` is set.
        If neither case, `subprocess` will be used.

        If `self.options.workDir` is defined, this path will be used as the
        temporary directory within the containers.

        If `self.options.volumes` is available, these will be mounted
        inside the containers. `volumes` must be a list of tuples:

            [(<local_path>, <container_absolute_path>), ...]

        Arguments:
            args (list): list of command line arguments passed to the tool.
            cwd (str): current working directory.
            env (dict): environment variables to set inside container.
            check_output (bool): if true, returns stdout of system call.

        Returns:
            str: (check_output=True) stdout of the system call.
            int: (check_output=False) 0 if call succeed else raise error.

        Raises:
            toil_container.UsageError: if both `singularity` and
                `docker` are set in `self.options`. Or if invalid
                `volumes` are defined.

            toil_container.SystemCallError: if system call cannot be completed.
        """
        call_kwargs = dict(args=args, env=env, cwd=cwd)
        docker = getattr(self.options, "docker", None)
        singularity = getattr(self.options, "singularity", None)

        if singularity and docker:
            raise exceptions.UsageError(
                "Both docker and singularity can't be set "
                "at the same time."
                )

        if singularity or docker:
            call_kwargs["check_output"] = check_output

            # used for testing only
            call_kwargs["remove_tmp_dir"] = getattr(self, "_rm_tmp_dir", True)

            if getattr(self.options, "workDir", None):
                call_kwargs["working_dir"] = self.options.workDir

            if getattr(self.options, "volumes", None):
                call_kwargs["volumes"] = self.options.volumes

            if singularity:
                call_kwargs["image"] = singularity
                call_function = containers.singularity_call
            else:
                call_kwargs["image"] = docker
                call_function = containers.docker_call

        elif check_output:
            call_function = subprocess.check_output

        else:
            call_function = subprocess.check_call

        errors = (
            exceptions.ContainerError,
            subprocess.CalledProcessError,
            OSError,
            )

        try:
            output = call_function(**call_kwargs)
        except errors as error:  # pylint: disable=catching-non-exception
            raise exceptions.SystemCallError(error)

        try:
            return output.decode()
        except AttributeError:
            return output
